"""
monobit.storage.fontfiles - load and save fonts

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import os
import sys
import logging
import shlex
from pathlib import Path
from contextlib import contextmanager

from ..constants import MONOBIT
from ..core import Font, Pack
from ..plumbing import scriptable, manage_arguments
from ..base import Any, FileFormatError, UnsupportedError
from .magic import MagicRegistry
from .location import open_location, iter_funcs_from_registry
from .base import (
    DEFAULT_TEXT_FORMAT, DEFAULT_BINARY_FORMAT,
    loaders, savers, container_loaders, container_savers
)


##############################################################################
# loading

@scriptable(wrapper=True, record=False)
def load(infile:Any='', *, format:str='', container_format:str='', match_case:bool=False, **kwargs):
    """
    Read font(s) from file.

    infile: input file or path (default: stdin)
    format: input format (default: infer from magic number or filename)
    container_format: container/wrapper formats separated by . (default: infer from magic number or filename)
    match_case: interpret path as case-sensitive (if file system supports it; default: False)
    """
    infile = infile or sys.stdin
    with open_location(
            infile, mode='r', match_case=match_case,
            container_format=container_format, argdict=kwargs,
        ) as location:
        if location.is_dir():
            return _load_container(
                location, format=format, **location.argdict
            )
        else:
            return _load_stream(
                location.get_stream(), format=format, **location.argdict
            )


def _load_stream(instream, *, format='', **kwargs):
    """Load fonts from open stream."""
    tried_formats = []
    for loader in iter_funcs_from_registry(loaders, instream, format):
        tried_formats.append(loader.format)
        instream.seek(0)
        logging.info("Loading '%s' as format `%s`", instream.name, loader.format)
        try:
            fonts = loader(instream, **kwargs)
        except FileFormatError as e:
            logging.debug(e)
        else:
            if fonts:
                break
            logging.debug(
                "No fonts found in '%s' as format `%s`.",
                instream.name, loader.format
            )
    else:
        message = f"Unable to read fonts from '{instream.name}': "
        if not tried_formats:
            message += f'format specifier `{format}` not recognised.'
        else:
            message += 'tried formats: ' + ', '.join(tried_formats)
        raise FileFormatError(message)
    # convert font or pack to pack
    pack = _annotate_fonts_with_source(
        fonts, instream.name, instream.where, loader.format, kwargs
    )
    return pack


def _sanitise_filesystem_name(filename):
    """
    If the source filename contains surrogate-escaped non-utf8 bytes
    preserve the byte values as backslash escapes
    """
    try:
        filename.encode('utf-8')
    except UnicodeError:
        filename = (
            filename.encode('utf-8', 'surrogateescape')
            .decode('ascii', 'backslashreplace')
        )
    return filename


def _annotate_fonts_with_source(
        fonts, filename, location, format, loader_kwargs
    ):
    """Set source metadata on font pack."""
    # convert font or pack to pack
    pack = Pack(fonts)
    filename = _sanitise_filesystem_name(Path(filename).name)
    filepath = _sanitise_filesystem_name(str(location.root) + str(location.path))
    # source format argumets
    loader_args = ' '.join(
        f'{_k.replace("_", "-")}={shlex.join((str(_v),))}'
        for _k, _v in loader_kwargs.items()
    )
    loader_args = f' [{loader_args}]' if loader_args else ''
    return Pack(
        _font.modify(
            converter=MONOBIT,
            source_format=_font.source_format or f'{format}{loader_args}',
            source_name=_font.source_name or filename,
            source_path=_font.source_path or filepath,
        )
        for _font in pack
    )


def _load_container(location, *, format='', **kwargs):
    """Load from a container."""
    for loader in iter_funcs_from_registry(
            container_loaders, instream=None, format=format
        ):
        logging.info("Loading '%s' as container format `%s`", location.path, loader.format)
        try:
            fonts = loader(location, **kwargs)
        except FileFormatError as e:
            logging.debug(e)
            continue
        spec_msg = f"as format {loader.format}"
        pack = _annotate_fonts_with_source(
            fonts, location.path, location, loader.format, kwargs
        )
        break
    else:
        pack = load_all(location, format=format, **kwargs)
        spec_msg = 'all'
    if not pack:
        raise FileFormatError(
            f"No fonts found in '{location.path}' while loading {spec_msg}."
        )
    return pack


def load_all(root_location, *, format='', **kwargs):
    """Open container and load all fonts found in it into one pack."""
    logging.info('Reading all from `%s`.', root_location)
    packs = Pack()
    for location in root_location.walk():
        with location:
            logging.debug('Trying `%s`.', location)
            try:
                pack = _load_stream(
                    location.get_stream(), format=format, **kwargs
                )
            except (FileFormatError, UnsupportedError) as exc:
                logging.debug('Could not load `%s`: %s', location, exc)
            else:
                packs += Pack(pack)
    if not packs:
        raise FileFormatError('Unable to read fonts from container.')
    return packs


##############################################################################
# saving

@scriptable(wrapper=True, record=False, pack_operation=True)
def save(
        pack_or_font,
        outfile:Any='', *,
        format:str='', overwrite:bool=False,
        container_format:str='',
        **kwargs
    ):
    """
    Write font(s) to file.

    outfile: output file or path (default: stdout)
    format: font file format (default: infer from filename)
    container_format: container/wrapper formats separated by . (default: infer from filename)
    overwrite: if outfile is a path, allow overwriting existing file
    """
    pack = Pack(pack_or_font)
    outfile = outfile or sys.stdout
    if outfile == sys.stdout:
        # errors can occur if the strings we write contain surrogates
        # these may come from filesystem names using 'surrogateescape'
        sys.stdout.reconfigure(errors='replace')
    if not pack:
        raise ValueError('No fonts to save')
    make_dir = format in container_savers.get_formats()
    with open_location(
            outfile, mode='w', overwrite=overwrite,
            container_format=container_format,
            argdict=kwargs,
            make_dir=make_dir,
        ) as location:
        if location.is_dir():
            _save_container(
                pack, location, format=format, **location.argdict
            )
        else:
            _save_stream(
                pack, location.get_stream(), format=format, **location.argdict
            )
    return pack_or_font


def _save_stream(pack, outstream, *, format='', **kwargs):
    """Save fonts to an open stream."""
    matching_savers = savers.get_for(outstream, format=format)
    if not matching_savers:
        if format:
            raise ValueError(f'Format specification `{format}` not recognised')
        else:
            raise ValueError(
                f'Could not infer output file format from filename `{outstream.name}`, '
                'please specify -format'
            )
    if len(matching_savers) > 1:
        raise ValueError(
            f"Format for output filename '{outstream.name}' is ambiguous: "
            f'specify -format with one of the values '
            f'({", ".join(_s.format for _s in matching_savers)})'
        )
    saver, *_ = matching_savers
    logging.info('Saving `%s` as %s.', outstream.name, saver.format)
    # apply wrappers to saver function
    saver = manage_arguments(saver)
    saver(pack, outstream, **kwargs)


def _save_container(pack, location, *, format, **kwargs):
    """Save font(s) to container."""
    for saver in iter_funcs_from_registry(
            container_savers, instream=None, format=format
        ):
        logging.info(
            "Saving '%s' as container format `%s`", location.path, saver.format
        )
        saver(pack, location, **kwargs)
        break
    else:
        save_all(pack, location, format=format, **kwargs)


def save_all(
        pack, location, *, format='', template='', **kwargs
    ):
    """Save fonts to a container."""
    format = format or DEFAULT_TEXT_FORMAT
    logging.info('Writing all to `%s`.', location)
    for font in pack:
        if format and not template:
            # generate name from format
            template = savers.get_template(format)
        # fill out template
        name = font.format_properties(template)
        # generate unique filename
        filename = location.unused_name(name.replace(' ', '_'))
        try:
            with location.join(filename) as new_location:
                _save_stream(
                    Pack(font),
                    new_location.get_stream(),
                    format=format,
                    **kwargs
                )
        except (FileFormatError, UnsupportedError) as e:
            logging.error('Could not save `%s`: %s', filename, e)



"""
monobit.encoding.registry - encoding registry

(c) 2020--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from functools import partial

from .base import Encoder, EncodingName, NotFoundError
from .charmaps import Charmap, Unicode


class EncodingRegistry:
    """Register and retrieve charmaps."""

    def __init__(self):
        self._index = {}
        self._encoders = []

    def __setitem__(self, names, encoder_or_callable):
        """Register an encoder to one or more aliases."""
        if isinstance(names, str):
            aliases = (names,)
        else:
            aliases = names
        for name in aliases:
            normname = EncodingName(name)
            if normname in self._index:
                logging.warning(f"Redefining encoder '{normname}'")
            self._index[normname] = len(self._encoders)
        self._encoders.append(encoder_or_callable)

    def _get_index(self, name):
        """Get index from registry by name; raise NotFoundError if not found."""
        normname = EncodingName(name)
        try:
            return self._index[normname]
        except KeyError as exc:
            raise NotFoundError(
                f"No registered character map matches '{name}' ['{normname}']."
            ) from exc

    def getter(self, name):
        """Get charmap or builder from registry by name; raise NotFoundError if not found."""
        index = self._get_index(name)
        return self._encoders[index]

    def __getitem__(self, name):
        index = self._get_index(name)
        encoder = self._encoders[index]
        """Get charmap from registry by name; raise NotFoundError if not found."""
        # must be either an Encoder or a callable that produces one
        if not isinstance(encoder, Encoder):
            encoder = encoder()
            self._encoders[index] = encoder
        return encoder

    ###

    def is_unicode(self, name):
        """Encoding name is equivalent to unicode."""
        try:
            return isinstance(self[name], Unicode)
        except NotFoundError:
            return False

    def __iter__(self):
        """Iterate over names of registered charmaps."""
        return iter(self._index.keys())

    def fit(self, charmap):
        """Return best-fit registered charmap."""
        min_dist = len(charmap)
        fit = Charmap()
        for registered in self:
            registered_map = self[registered]
            if not isinstance(registered_map, Charmap):
                continue
            dist = charmap.distance(registered_map)
            if dist == 0:
                return registered_map
            elif dist < min_dist:
                min_dist = dist
                fit = registered_map
        return fit



"""
monobit.storage.fontformats.apple - Mac OS and other NFNT-based font formats

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from monobit.base import import_all
import_all(__name__)



"""
monobit.storage.fontformats.image.image - fonts stored in image files

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from collections import Counter
from pathlib import Path

from monobit.base import safe_import
Image = safe_import('PIL.Image')

from monobit.base import Coord, RGB, FileFormatError, UnsupportedError
from monobit.base.binary import ceildiv
from monobit.storage.base import (
    loaders, savers, container_loaders, container_savers
)
from monobit.core import Font, Glyph, Codepoint
from monobit.render import create_chart, glyph_to_image, grid_traverser
from monobit.storage.utils.limitations import ensure_single
from monobit.storage.utils.perglyph import loop_load, loop_save


DEFAULT_IMAGE_FORMAT = 'png'


# available background policies
# -----------------------------
#
# most-common       use colour most commonly found in payload cells
# least-common      use colour least commonly found in payload cells
# brightest         use brightest colour, by sum of RGB values
# darkest           use darkest colour, by sum of RGB values
# top-left          use colour of top-left pixel in first cell

def identify_inklevels(colours, background):
    """Identify ink levels from colour set."""
    colourset = set(colours)
    if len(colourset) < 2:
        raise FileFormatError('No glyphs or only blank glyphs found.')
    elif len(colourset) > 2:
        # 3 or more non-border colours, must be a greyscale image
        if not all(
                len(set(_c[:3])) == 1 and not _c[3:] or _c[3] == 255
                for _c in colourset
            ):
            # only greyscale allowed, r==g==b, alpha==255
            raise UnsupportedError('Colour fonts not supported.')
        # get a random element to check colour mode (8/24/32 bit)
        tuple_len = len(colourset.pop())
        if tuple_len == 4:
            # RGBA
            inklevels = tuple((_c, _c, _c, 255) for _c in range(256))
        else:
            # RGB or 8-bit
            inklevels = tuple((_c,) * tuple_len for _c in range(256))
        return inklevels
    else:
        # 2-colour image
        if not isinstance(background, str):
            # background provided
            paper = background
        elif background in ('most-common', 'least-common'):
            colourfreq = Counter(colours)
            if background == 'most-common':
                # most common colour in image assumed to be background colour
                paper, _ = colourfreq.most_common(1)[0]
            else:
                # least common colour in image assumed to be background colour
                paper, _ = colourfreq.most_common()[-1]
        elif background in ('darkest', 'brightest'):
            brightness = sorted((sum(_c), _c) for _c in colourset)
            if background == 'darkest':
                # darkest colour assumed to be background
                _, paper = brightness[0]
            else:
                # brightest colour assumed to be background
                _, paper = brightness[-1]
        elif background == 'top-left':
            # top-left pixel of first char assumed to be background colour
            paper = colours[0]
        else:
            raise ValueError(f'Background mode `{background}` not supported.')
        # 2 colour image - not-paper means ink
        ink = (colourset - {paper}).pop()
        return paper, ink


if Image:
    @loaders.register(
        name='image',
        patterns=(
            '*.png', '*.bmp', '*.gif', '*.tif', '*.tiff',
            '*.ppm', '*.pgm', '*.pbm', '*.pnm', '*.webp',
            '*.pcx', '*.tga', '*.jpg', '*.jpeg',
        ),
        magic=(
            # PNG
            b'\x89PNG\r\n\x1a\n',
            # BMP
            #b'BM',   # -- clash with bmfont b'BMF'
            # GIF
            b'GIF87a', b'GIF89a',
            # TIFF
            b'\x4D\x4D\x00\x2A', b'\x49\x49\x2A\x00'
            # PNM
            b'P1', b'P2', b'P3',
            # WebP
            b'RIFF',
            # PCX
            b'\n\x00', b'\n\x02', b'\n\x03', b'\n\x04', b'\n\x05',
            # JPEG
            b'\xFF\xD8\xFF',
        ),
    )
    def load_image(
            infile,
            cell:Coord=None,
            margin:Coord=Coord(0, 0),
            padding:Coord=Coord(1, 1),
            scale:Coord=Coord(1, 1),
            table_size:Coord=None,
            count:int=0,
            background:str='most-common',
            first_codepoint:int=0,
            direction:str='left-to-right top-to-bottom',
            keep_empty:bool=False,
            grid:bool=False,
        ):
        """
        Extract font from image.

        grid: extract on a rigid grid (default: False)
        cell: (grid) glyph raster size X,Y. 0 or negative: calculate from table_size (default)
        margin: (grid) number of pixels in X,Y direction around glyph chart (default: 0x0)
        padding: (grid) number of pixels in X,Y direction between glyph (default: 1x1)
        table_size: (grid) number of glyphs in X, Y direction. 0 or negative means as much as fits on the axis. (default: 32x8).
        scale: number of pixels in X,Y direction per glyph bit (default: 1x1)
        count: maximum number of glyphs to extract (within constraints of table_size). 0 or negative means extract all (default).
        background: determine background from "most-common" (default), "least-common", "brightest", "darkest", "top-left" colour
        first_codepoint: codepoint value assigned to first glyph (default: 0)
        direction: two-part string, default 'left-to-right top-to-bottom'
        keep_empty: keep empty glyphs (default: False)
        """
        with Image.open(infile) as img:
            img = img.convert('RGB')
            if grid:
                crops = extract_crops_from_grid(
                    img, table_size, cell, scale, padding, margin, direction
                )
            else:
                crops = extract_crops_from_strips(img, direction)
        if not crops:
            logging.error('Could not extract glyphs from image.')
            return Font()
        # scale
        crops = tuple(
            _crop.resize(
                (_crop.width // scale.x, _crop.height // scale.y),
                resample=Image.NEAREST,
            )
            for _crop in crops
        )
        if count > 0:
            crops = crops[:count]
        return convert_crops_to_font(
            enumerate(crops, first_codepoint), background, keep_empty
        )


    def extract_crops_from_grid(
            img, table_size, cell, scale, padding, margin, direction
        ):
        """Extract glyph crops from grid-based image."""
        # appply defaults
        if table_size is None:
            if cell is None:
                table_size = Coord(32, 8)
                cell = Coord(0, 0)
            else:
                table_size = Coord(0, 0)
        elif cell is None:
            cell = Coord(0, 0)
        (
            cell, step_x, step_y, table_size_x, table_size_y
        ) = determine_grid_geometry(
            img.width, img.height, table_size, cell, scale, padding, margin,
        )
        traverse = grid_traverser(table_size_x, table_size_y, direction)
        # extract sub-images
        crops = tuple(
            img.crop((
                min(img.width, margin.x + _col*step_x),
                max(0, img.height - (margin.y + _row*step_y + cell.y*scale.y)),
                min(img.width, margin.x + _col*step_x + cell.x*scale.x),
                max(0, img.height - (margin.y + _row*step_y)),
            ))
            for _row, _col in traverse
        )
        # three-colour mode - proportional width encoded with border colour
        colourset = set(img.getdata())
        if len(colourset) >= 3:
            # get border/padding colour
            border = _get_border_colour(img, cell, margin, padding)
            # clip off border colour from cells
            crops = tuple(_crop_border(_crop, border) for _crop in crops)
        return crops


    def determine_grid_geometry(
            width, height, table_size, cell, scale, padding, margin,
        ):
        """Find cell, step, row and column sizes."""
        # determine defaults & whether to work with cell-size or table size
        # maximum number of cells that fits
        cell_x, cell_y = cell
        if cell.x <= 0:
            if table_size.x <= 0:
                raise ValueError('Either cell or table size must be specified.')
            cell_x = ceildiv(width, table_size.x*scale.x) - padding.x
        if cell.y <= 0:
            if table_size.y <= 0:
                raise ValueError('Either cell or table size must be specified.')
            cell_y = ceildiv(height, table_size.y*scale.y) - padding.y
        if not cell_x or not cell_y:
            raise ValueError('Empty cell. Please specify larger cell size or smaller table size.')
        logging.debug('Cell size %dx%d', cell_x, cell_y)
        cell = Coord(cell_x, cell_y)
        # work out image geometry
        step_x = cell.x * scale.x + padding.x
        step_y = cell.y * scale.y + padding.y
        table_size_x, table_size_y = table_size
        if table_size.x <= 0:
            table_size_x = ceildiv(width - margin.x, step_x)
        if table_size.y <= 0:
            table_size_y = ceildiv(height - margin.y, step_y)
        return cell, step_x, step_y, table_size_x, table_size_y


    def convert_crops_to_font(enumerated_crops, background, keep_empty):
        """Convert list of glyph images to font."""
        enumerated_crops = tuple(enumerated_crops)
        # get pixels
        _, crops = tuple(zip(*enumerated_crops))
        inklevels = _identify_colours(crops, background)
        # convert to glyphs, set codepoints
        glyphs = tuple(
            Glyph.from_vector(
                tuple(_crop.getdata()),
                stride=_crop.width,
                inklevels=inklevels,
                codepoint=_index,
            )
            for _index, _crop in enumerated_crops
        )
        # drop empty glyphs
        if not keep_empty:
            glyphs = tuple(_g for _g in glyphs if _g.height and _g.width)
        return Font(glyphs)

    def _get_border_colour(img, cell, margin, padding):
        """Get border/padding colour."""
        if margin.x or margin.y:
            return img.getpixel((0, 0))
        elif padding.x:
            return img.getpixel((cell.x, 0))
        elif padding.y:
            return img.getpixel((0, cell.y))
        # can't determine border colour without padding or margin
        return None

    def _identify_colours(crops, background):
        """Identify ink levels from cells."""
        crops = (tuple(_crop.getdata()) for _crop in crops)
        colours = sum(crops, ())
        return identify_inklevels(colours, background)

    def _crop_border(image, border):
        """Remove border area from image."""
        if border is None:
            return image
        while image.width:
            right_colours = image.crop((
                image.width-1, 0, image.width, image.height
            )).getcolors()
            if len(right_colours) == 1 and right_colours[0][1] == border:
                image = image.crop((0, 0, image.width-1, image.height))
            else:
                break
        return image


    def extract_crops_from_strips(img, direction):
        """Extract glyph crops from strip-based image."""
        # we extract left-to-right or top-to-bottom
        glyph_dir, _, line_dir = direction.lower().partition(' ')
        glyph_dir = glyph_dir[:1] or 'l'
        line_dir = line_dir[:1] or 't'
        vertical = glyph_dir in ('t', 'b')
        strips, border = chop_strips(img, border=None, vertical=not vertical)
        if glyph_dir in ('r', 'b'):
            strips = strips[::-1]
        crops = []
        for strip in strips:
            strip_crops, _ = chop_strips(strip, border, vertical=vertical)
            if line_dir in ('r', 'b'):
                crops = crops[::-1]
            crops.extend(strip_crops)
        return crops


    def chop_strips(img, border, vertical):
        """Slice up image into strips by border colour."""
        if vertical:
            scan_range = range(img.height)
            def _get_slice(start, stop):
                return (0, start, img.width, stop)
        else:
            scan_range = range(img.width)
            def _get_slice(start, stop):
                return (start, 0, stop, img.height)
        strips = []
        last_line = 0
        for i_line in scan_range:
            line = img.crop(_get_slice(i_line, i_line+1))
            colours = line.getcolors()
            # identify border colour
            # the first full row of one colour is deemed to be border
            if len(colours) == 1 and border is None or colours[0][1] == border:
                # found full row of one colour
                if border is None:
                    border = colours[0][1]
                if i_line - last_line:
                    strip = img.crop(_get_slice(last_line, i_line))
                    strips.append(strip)
                last_line = i_line + 1
        if i_line + 1 - last_line:
            strip = img.crop(_get_slice(last_line, i_line+1))
            strips.append(strip)
        return strips, border


    ###########################################################################

    @savers.register(linked=load_image)
    def save_image(
            fonts, outfile, *,
            image_format:str='png',
            glyphs_per_line:int=32,
            margin:Coord=Coord(0, 0),
            padding:Coord=Coord(1, 1),
            scale:Coord=Coord(1, 1),
            direction:str='left-to-right top-to-bottom',
            border:RGB=RGB(32, 32, 32),
            paper:RGB=RGB(0, 0, 0),
            ink:RGB=RGB(255, 255, 255),
            codepoint_range:tuple[Codepoint]=None,
            grid_positioning:bool=True,
        ):
        """
        Export font to grid-based image.

        image_format: image file format (default: png)
        glyphs_per_line: number of glyphs per line in glyph chart (default: 32)
        margin: number of pixels in X,Y direction around glyph grid (default: 0x0)
        padding: number of pixels in X,Y direction between glyphs (default: 1x1)
        scale: number of pixels in X,Y direction per glyph bit (default: 1x1)
        direction: two-part string, default 'left-to-right top-to-bottom'
        paper: background colour R,G,B 0--255 (default: 0,0,0)
        ink: full-intensity foreground colour R,G,B 0--255 (default: 255,255,255)
        border: border colour R,G,B 0--255 (default 32,32,32)
        codepoint_range: range of codepoints to include (includes bounds and undefined codepoints; default: all codepoints)
        grid_positioning: place codepoints on corresponding grid positions, leaving gaps if undefined (default: false)
        """
        glyph_map = create_chart(
            fonts,
            glyphs_per_line=glyphs_per_line,
            margin=margin,
            padding=padding,
            scale=scale,
            direction=direction,
            codepoint_range=codepoint_range,
            grid_positioning=grid_positioning,
        )
        img, = glyph_map.to_images(
            border=border, paper=paper, ink=ink, transparent=False
        )
        try:
            img.save(outfile, format=image_format or Path(outfile).suffix[1:])
        except (KeyError, ValueError, TypeError):
            img.save(outfile, format=DEFAULT_IMAGE_FORMAT)


    ###########################################################################
    # image-set

    @container_loaders.register(name='imageset')
    def load_imageset(
            location,
            background:str='most-common',
            prefix:str='',
            base:int=16,
        ):
        """
        Extract font from per-glyph images.

        background: determine background from "most-common" (default), "least-common", "brightest", "darkest", "top-left" colour
        prefix: part of the image file name before the codepoint
        base: radix of numerals in file name representing code point
        """
        def _load_image_glyph(stream):
            crop = Image.open(stream)
            crop = crop.convert('RGB')
            cp = int(Path(stream.name).stem.removeprefix(prefix), base)
            # return codepoint, image pair to be parsed by convert_crops_to_font
            return (cp, crop)

        crops = loop_load(location, _load_image_glyph)
        return convert_crops_to_font(crops, background, keep_empty=True)


    @container_savers.register(linked=load_imageset)
    def save_imageset(
            fonts, location,
            prefix:str='',
            image_format:str='png',
            paper:RGB=(0, 0, 0),
            ink:RGB=(255, 255, 255),
        ):
        """
        Export font to per-glyph images.

        prefix: part of the image file name before the codepoint
        image_format: image file format (default: png)
        paper: background colour R,G,B 0--255 (default: 0,0,0)
        ink: foreground colour R,G,B 0--255 (default: 255,255,255)
        """
        def _save_image_glyph(glyph, imgfile):
            img = glyph_to_image(glyph, paper, ink)
            try:
                img.save(imgfile, format=image_format or Path(imgfile).suffix[1:])
            except (KeyError, ValueError, TypeError):
                img.save(imgfile, format=DEFAULT_IMAGE_FORMAT)

        loop_save(
            fonts, location, prefix,
            suffix=image_format, save_func=_save_image_glyph
        )



"""
monobit.storage.containerformats.directory - directory traversal

(c) 2021--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import os
import io
import logging
from pathlib import Path
from collections import deque

from ..streams import Stream
from ..base import containers
from ..containers import Container


@containers.register(name='dir')
class Directory(Container):
    """Treat directory tree as a container."""

    def __init__(self, path='', mode='r'):
        """Create directory wrapper."""
        # if empty path, this refers to the whole filesystem
        if not path:
            self._path = Path('/')
        elif isinstance(path, Directory):
            self._path = Path(path._path)
        else:
            self._path = Path(path)
        # mode really should just be 'r' or 'w'
        mode = mode[:1]
        if mode == 'w':
            logging.debug('Creating directory `%s`', self._path)
            # exist_ok raises FileExistsError only if the *target* already
            # exists, not the parents
            self._path.mkdir(parents=True, exist_ok=True)
        super().__init__(mode, str(self._path))

    def __repr__(self):
        return f"{type(self).__name__}('{self._path}')"

    def __str__(self):
        return str(self._path)

    def decode(self, name):
        return self.open(name, mode='r')

    def encode(self, name):
        return self.open(name, mode='w')

    def open(self, name, mode):
        """Open a stream in the directory."""
        # mode in 'r', 'w'
        mode = mode[:1]
        pathname = Path(name)
        if mode == 'w':
            path = pathname.parent
            logging.debug("Creating directory '%s'", self._path / path)
            (self._path / path).mkdir(parents=True, exist_ok=True)
        logging.debug("Opening file '%s' for mode '%s'.", name, mode)
        filepath = self._path / pathname
        file = open(filepath, mode + 'b')
        # provide name relative to directory container
        stream = Stream(file, name=str(pathname), mode=mode)
        return stream

    def remove(self, name):
        """Remove a file from the directory."""
        if not name:
            return
        filepath = self._path / name
        if not filepath.is_dir():
            filepath.unlink()

    def is_dir(self, name):
        """Item at `name` is a directory."""
        filepath = self._path / name
        return filepath.is_dir()

    def iter_sub(self, prefix):
        """List contents of a subpath."""
        return (
            _path.relative_to(self._path)
            for _path in (self._path / prefix).iterdir()
        )



"""
monobit.storage.utils.perglyph - utilities for one-glyph-per-file formats

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.storage.utils.limitations import ensure_single


def loop_save(fonts, location, prefix, suffix, save_func):
    """Loop over per-glyph files in container."""
    font = ensure_single(fonts)
    font = font.label(codepoint_from=font.encoding)
    font = font.equalise_horizontal()
    width = len(f'{int(max(font.get_codepoints())):x}')
    for glyph in font.glyphs:
        if not glyph.codepoint:
            logging.warning('Cannot store glyph without codepoint label.')
            continue
        cp = f'{int(glyph.codepoint):x}'.zfill(width)
        name = f'{prefix}{cp}.{suffix}'
        with location.open(name, 'w') as imgfile:
            save_func(glyph, imgfile)


def loop_load(location, load_func):
    """Loop over per-glyph files in container."""
    glyphs = []
    for name in sorted(location.iter_sub('')):
        with location.open(name, mode='r') as stream:
            glyphs.append(load_func(stream))
    return glyphs



"""
monobit.storage.fontformats.text.draw - hexdraw format

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import string

from monobit.storage.base import loaders, savers
from monobit.core import Font, Glyph, Tag, Char
from monobit.encoding import encodings
from monobit.base.binary import align

from monobit.storage.utils.limitations import ensure_single


##############################################################################
# hexdraw

@loaders.register(
    name='hexdraw',
    patterns=('*.draw',),
    text=True,
)
def load_hexdraw(
        instream, ink:str='#', paper:str='-', unicode:bool=True, base:int=16,
    ):
    """
    Load font from a hexdraw or similar text-based file.

    ink: character used for inked/foreground pixels (default #)
    paper: character used for uninked/background pixels (default -)
    unicode: interpret codepoint as Unicode (default: True)
    base: representational base for codepoint (default: 16 for hexadecimal)
    """
    class CustomGlyph(DrawGlyph): pass
    CustomGlyph.ink = ink
    CustomGlyph.paper = paper
    CustomGlyph.base = base
    return load_draw(
        instream.text, blocktypes=(CustomGlyph, DrawComment, Empty),
        unicode=unicode
    )


@savers.register(linked=load_hexdraw)
def save_hexdraw(fonts, outstream, ink:str='#', paper:str='-', unicode:bool=True):
    """
    Save font to a hexdraw file.

    ink: character to use for inked/foreground pixels (default #)
    paper: character to use for uninked/background pixels (default -)
    unicode: use unicode char labels for codepoints (default: True)
    """
    font = ensure_single(fonts)
    _save_draw(font, outstream.text, ink=ink, paper=paper, unicode=unicode)


# read hexdraw file

def load_draw(text_stream, *, blocktypes, unicode):
    """Parse a hexdraw-style file."""
    glyphs = []
    font_comments = []
    current_comment = []
    for block in iter_blocks(text_stream, blocktypes):
        if isinstance(block, DrawComment):
            if not glyphs:
                font_comments.extend(current_comment)
                current_comment = []
            current_comment.append(block.get_value())
        elif isinstance(block, DrawGlyph):
            if not glyphs and not font_comments:
                font_comments.extend(current_comment)
                current_comment = []
            glyphs.append(block.get_value().modify(
                comment='\n\n'.join(current_comment),
            ))
            current_comment = []
        elif isinstance(block, Unparsed):
            logging.debug('Unparsed lines: %s', block.get_value())
    font_comments.extend(current_comment)
    if unicode:
        encoding = 'unicode'
    else:
        encoding = ''
        glyphs = tuple(
            _g.label(codepoint_from=encodings['unicode']).modify(char=None)
            for _g in glyphs
        )
    font = Font(
        glyphs,
        encoding=encoding,
        comment='\n\n'.join(font_comments)
    )
    return font


# write hexdraw file

def _save_draw(font, outstream, *, ink, paper, unicode):
    """Write one font to a plaintext stream as hexdraw."""
    font = font.equalise_horizontal()
    if unicode:
        # ensure char labels are set
        font = font.label(char_from=font.encoding, match_whitespace=False, match_graphical=False)
    # write global comment
    if font.get_comment():
        outstream.write(
            format_comment(font.get_comment(), comment_char='%') + '\n'
        )
    # write glyphs
    for i, glyph in enumerate(font.glyphs):
        if unicode:
            char = glyph.char
        else:
            cpbytes = bytes(glyph.codepoint)
            cpbytes = cpbytes.rjust(align(len(cpbytes), 2), b'\0')
            char = cpbytes.decode('utf-32-be')
        if not char:
            logging.warning(
                "Can't encode glyph without Unicode character label in .draw file;"
                " skipping index %d", i
            )
        elif len(char) > 1:
            logging.warning(
                "Can't encode grapheme cluster %s in .draw file; skipping.",
                ascii(char)
            )
        else:
            glyphtxt = glyph.as_text(
                start='\t', inklevels=(paper, ink), end='\n'
            )
            outstream.write(f'\n{ord(char):04x}:')
            outstream.write(glyphtxt)


def format_comment(comments, comment_char):
    """Format a multiline comment."""
    return '\n'.join(
        f'{comment_char} {_line}'
        for _line in comments.splitlines()
    )


##############################################################################
# common utilities and reader classes

def iter_blocks(text_stream, classes):
    block = BlockBuilder(classes)
    for line in text_stream:
        while not block.append(line.rstrip('\r\n')):
            yield block.emit()
            block = BlockBuilder(classes)
    if block:
        yield block.emit()


class BlockBuilder:

    def __init__(self, classes):
        self.block = None
        self.classes = (*classes, Unparsed)

    def __bool__(self):
        return self.block is not None

    def append(self, line):
        if not self.block:
            for blocktype in self.classes:
                try:
                    self.block = blocktype(line)
                    return True
                except ValueError:
                    pass
            else:
                # this should not happen - BaseBlock absorbs
                raise ValueError('unparsed block')
        if self.block.ends(line):
            return False
        self.block.append(line)
        return True

    def emit(self):
        return self.block


class BaseBlock:

    def __init__(self, line):
        if not self.starts(line):
            raise ValueError(f'{line:r} cannot start a {type(self).__name__}')
        self.lines = []
        self.append(line)

    def __repr__(self):
        return f'{type(self).__name__}({repr(self.lines)})'

    def starts(self, line):
        return True

    def ends(self, line):
        return True

    def append(self, line):
        self.lines.append(line)

    def get_value(self):
        return tuple(self.lines)


class Unparsed(BaseBlock):
    pass


class Empty(BaseBlock):

    def starts(self, line):
        return not line


class NonEmptyBlock(BaseBlock):

    def append(self, line):
        line = line.rstrip()
        if line:
            self.lines.append(line)


# draw format block readers


class DrawComment(NonEmptyBlock):

    notcomment = string.hexdigits + string.whitespace

    def starts(self, line):
        return line[:1] not in self.notcomment

    def ends(self, line):
        return not self.starts(line)

    def get_value(self):
        lines = tuple(self.lines)
        first = equal_firsts(lines)
        if first and first not in string.ascii_letters + string.digits:
            lines = tuple(_line[1:] for _line in lines)
        if equal_firsts(lines) == ' ':
            lines = (_line[1:] for _line in lines)
        return '\n'.join(lines)

def equal_firsts(lines):
    first_chars = set(_line[:1] for _line in lines if _line)
    if len(first_chars) == 1:
        return first_chars.pop()
    return None


class DrawGlyph(NonEmptyBlock):
    paper = '-'
    ink = '#'

    # base for codepoint. default: hexadecimal
    base = 16

    def starts(self, line):
        return line and line[:1] in string.hexdigits

    def ends(self, line):
        # empty or non-indented
        return not line[:1] in string.whitespace

    def get_value(self):
        key, _, value =  self.lines[0].partition(':')
        value = value.strip()
        lines = self.lines[1:]
        if value:
            lines = [value] + lines
        lines = tuple(_l.strip() for _l in lines)
        return Glyph(
            lines, inklevels=(self.paper, self.ink),
            labels=(self.convert_key(key),),
        )

    @classmethod
    def convert_key(cls, key):
        """Convert keys on input from .draw."""
        key = key.strip()
        try:
            return Char(''.join(
                chr(int(_key, cls.base))
                for _key in key.split(',')
            ))
        except (TypeError, ValueError):
            return Tag(key)



"""
monobit.storage.utils.limitations - deal with font format limitations

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from monobit.base import Coord
from monobit.core import Glyph
from monobit.base import UnsupportedError


def ensure_single(fonts):
    font, *more = fonts
    if more:
        raise UnsupportedError('This format can only store one font per file.')
    return font


def ensure_charcell(font, cell_size=None):
    # check if font is fixed-width and fixed-height
    if font.spacing != 'character-cell':
        raise UnsupportedError(
            'This format only supports character-cell fonts.'
        )
    if cell_size and font.cell_size != cell_size:
        raise UnsupportedError(
            f'This format only supports {Coord(cell_size)} character-cell fonts.'
        )
    # fill out character cell including shifts, bearings and line height
    font = font.equalise_horizontal()
    return font


def make_contiguous(font, *, missing, supported_range=None, full_range=None):
    """Fill out a contiguous range of glyphs."""
    # fill in codepoints where possible
    font = font.label(codepoint_from=font.encoding)
    if not full_range:
        # we need a contiguous range between the min and max codepoints
        min_range = int(min(font.get_codepoints()))
        max_range = int(max(font.get_codepoints()))
        if supported_range:
            min_range = max(min_range, min(supported_range))
            max_range = min(max_range, max(supported_range))
        full_range = range(min_range, max_range+1)
    font = font.resample(codepoints=full_range, missing=missing)
    return font



"""
monobit.storage.fontformats.text.yaff - monobit-yaff format

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
import string
from dataclasses import dataclass, field
from itertools import count, zip_longest
from collections import deque
from functools import cached_property

from monobit.storage import loaders, savers
from monobit.storage.magic import Sentinel
from monobit.core import (
    Font, FontProperties, Glyph, Raster, Label, strip_matching, CUSTOM_NAMESPACE
)
from monobit.base.binary import INKLEVELS
from monobit.base import Props, Coord, passthrough, FileFormatError

from .draw import NonEmptyBlock, DrawComment, Empty, Unparsed, iter_blocks
from .draw import format_comment


##############################################################################
# interface

@loaders.register(
    name='yaff',
    magic=(
        # maybe, if multi-section
        b'---',
        # maybe
        Sentinel(b'yaff:'),
    ),
    patterns=('*.yaff', '*.yaffs',),
    text=True,
)
def load_yaff(instream, allow_empty:bool=False):
    """
    Load font from a monobit .yaff file.

    allow_empty: allow files with no glyphs (default: False)
    """
    return _load_yaffs(instream.text, allow_empty)


@savers.register(
    linked=load_yaff,
)
def save_yaff(fonts, outstream):
    """Write fonts to a monobit .yaff file."""
    _save_yaff(fonts, outstream.text)


##############################################################################
# format parameters

BOUNDARY_MARKER = '---'


class YaffParams:
    """Parameters for .yaff format."""

    # first/second pass constants
    separator = ':'
    comment = '#'
    # output only
    tab = '    '
    # tuple of individual chars, need to be separate for startswith
    whitespace = tuple(' \t')

    # third-pass constants
    ink = '@'
    paper = '.'
    empty = '-'

    # string to be quoted if one of these chars at start and/or end
    quotable = (':', ' ')
    glyphchars = (ink, paper, empty)

    inklevels = {
        256: ('..', *(f'{_c:02X}' for _c in range(1, 255)), '@@'),
        16: paper + INKLEVELS[16][1:-1].upper() + ink,
        4: paper + INKLEVELS[4][1:-1].upper() + ink,
        2: paper + ink,
    }


# deprecated compatibility synonymms
_DEPRECATED_SYNONYMS = {
    'kern_to': 'right_kerning',
    'tracking': 'right_bearing',
    'offset': ('left_bearing', 'shift_up'),

    'average_advance': 'average_width',
    'max_advance': 'max_width',
    'cap_advance': 'cap_width',
}
def _set_property(propsdict, key, value):
    try:
        key = _DEPRECATED_SYNONYMS[key]
    except KeyError:
        pass
    if isinstance(key, tuple):
        for key, value in zip(key, Coord.create(value)):
            propsdict[key] = value
    else:
        propsdict[key] = value



##############################################################################
# read file


def _load_yaffs(text_stream, allow_empty):
    """Parse a yaff or yaffs file."""
    fonts = []
    reader = SectionIterator(text_stream)
    while not reader.eof:
        font = _read_yaff(reader)
        # if no glyphs, ignore it - may not be yaff at all
        if font.glyphs or allow_empty:
            fonts.append(font)
    return fonts


class SectionIterator:

    def __init__(self, textstream):
        self._stream = textstream
        self.eof = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.eof:
            raise StopIteration()
        line = self._stream.readline()
        if not line:
            self.eof = True
            raise StopIteration()
        if line[:3] == BOUNDARY_MARKER:
            raise StopIteration()
        else:
            return line


def _read_yaff(text_stream):
    """Parse a monobit yaff file."""
    blocktypes = (
        YaffComment, YaffProperty, YaffGlyph, YaffPropertyOrGlyph,
        Empty, Unparsed
    )
    glyphs = []
    font_comments = []
    font_props = {}
    font_prop_comms = {}
    current_comment = []
    inklevels = YaffParams.inklevels[2]
    for block in iter_blocks(text_stream, blocktypes):
        if isinstance(block, (YaffGlyph, YaffPropertyOrGlyph)) and block.is_glyph():
            glyphs.append(block.get_glyph_value(inklevels) | Props(
                comment='\n\n'.join(current_comment),
            ))
            current_comment = []
        elif isinstance(block, (YaffProperty, YaffPropertyOrGlyph)):
            key = block.get_key()
            value = block.get_value()
            if key == 'yaff':
                logging.debug("yaff signature found, version %s", value)
            else:
                _set_property(font_props, key, value)
                if key == 'levels':
                    inklevels = YaffParams.inklevels[int(value, 0)]
            font_prop_comms[key] = '\n\n'.join(current_comment)
            current_comment = []
        if not glyphs and not font_props:
            font_comments.extend(current_comment)
            current_comment = []
        if isinstance(block, YaffComment):
            current_comment.append(block.get_value())
        elif not isinstance(block, (YaffProperty, YaffGlyph, YaffPropertyOrGlyph)):
            logging.debug('Unparsed lines: %s', block.get_value())
    font_comments.extend(current_comment)
    # construct glyphs, including path-only glyphs
    glyphs = (
        Glyph(**vars(_g)) if _g.pixels or not hasattr(_g, 'path')
        else Glyph.from_path(**vars(_g - 'pixels'))
        for _g in glyphs
    )
    return Font(
        glyphs, **font_props,
        comment={'': '\n\n'.join(font_comments), **font_prop_comms},
    )


class YaffComment(DrawComment):

    def starts(self, line):
        return line[:1] == YaffParams.comment


class YaffMultiline(NonEmptyBlock, YaffParams):

    def __init__(self, line):
        self.indent = 0
        super().__init__(line)

    def ends(self, line):
        if self.indent:
            # dedent
            return line and line[:1] not in self.whitespace
        if line and line[:1] in self.whitespace:
            return False
        # gather multiple labels without values, but break on line with value
        return line.rstrip()[-1:] != self.separator

    def append(self, line):
        line = line.rstrip()
        self.lines.append(line)
        if not self.indent and line and line[:1] in self.whitespace:
            self.indent = len(line) - len(line.lstrip())


class YaffGlyph(YaffMultiline):
    """Glyph with 1.0 label. Monochrome or greyscale."""

    label_chars = tuple('"' + "'" + string.digits)

    def starts(self, line):
        return line and (line[:1] in self.label_chars) or '+' in line

    def get_value(self, inklevels):
        labels = tuple(_l[:-1] for _l in self.lines[:self.n_keys])
        lines = (_l[self.indent:] for _l in self.lines[self.n_keys:])
        lines = tuple(_l for _l in lines if _l)
        # locate glyph properties
        i = 0
        for i, line in enumerate(lines):
            if self.separator in line:
                break
        else:
            i += 1
        if not i:
            raise FileFormatError('Malformed yaff file: expected glyph definition.')
        raster = tuple(_l.upper() for _l in lines[:i])
        properties = {}
        key = None
        for line in lines[i:]:
            if line[:1] in self.whitespace:
                # follow-up lines in multiline glyph properties
                # note - won't work with deprecated synonyms
                if not properties[key]:
                    properties[key] = line.strip()
                else:
                    properties[key] = '\n'.join((properties[key], line.strip()))
            else:
                # first line of property
                # one-line glyph properties
                key, _, value = line.partition(self.separator)
                key = normalise_property(key)
                value = value.strip()
                _set_property(properties, key, value)
        # deal with sized empties (why?)
        if all(set(_line) == set([self.empty]) for _line in raster):
            raster = Raster.blank(
                width=len(raster[0])-1, height=len(raster)-1,
                levels=len(inklevels),
            )
        # split up pixel groups for 8-bit
        elif len(inklevels) == 256:
            raster = Raster.from_matrix(
                (
                    tuple(_row[_c:_c+2] for _c in range(0, len(_row), 2))
                    for _row in raster
                ),
                inklevels=inklevels,
            )
        return Props(
            pixels=Raster(raster, inklevels=inklevels),
            labels=labels, **properties
        )

    def _count_keys(self):
        return sum(
            _line[-1:] == self.separator and _line[:1] not in self.whitespace
            for _line in self.lines
        )
    n_keys = cached_property(_count_keys)

    def is_glyph(self):
        return True

    get_glyph_value = get_value



def normalise_property(field):
    # preserve distinction between starting underscore (internal) and starting dash (user property)
    return field.replace('-', '_')


# keywords that take a label value
# these need special treatment as quotes must not be stripped
_LABEL_VALUED_KEYS = tuple(
    _k for _k, _v in FontProperties.__annotations__.items() if _v == Label
)


class YaffProperty(NonEmptyBlock, YaffParams):
    """Inline property."""

    def starts(self, line):
        return (
            line[:1] not in self.whitespace
            and line[-1:] != self.separator
            and self.separator in line
        )

    def get_value(self):
        key, _, value = self.lines[0].partition(self.separator)
        # label values need special treatment as quotes must not be stripped
        if key in _LABEL_VALUED_KEYS:
            return value
        return _strip_quotes(value)

    def get_key(self):
        key, _, _ = self.lines[0].partition(self.separator)
        return normalise_property(key)


def _strip_quotes(line):
    line = line.strip()
    if len(line) > 1 and line[0] == line[-1] == '"':
        return line[1:-1]
    return line


class YaffPropertyOrGlyph(YaffMultiline):
    """Multiline properties, or monochrome glyph with pre-1.0 tag/char label."""

    def starts(self, line):
        return line[:1] not in self.whitespace and line[-1:] == self.separator

    def get_value(self):
        return '\n'.join(_strip_quotes(_l) for _l in self.lines[1:])

    def get_key(self):
        return normalise_property(self.lines[0][:-1])

    # glyph block with plain label

    get_glyph_value = YaffGlyph.get_value
    n_keys = cached_property(YaffGlyph._count_keys)

    def is_glyph(self):
        if self.n_keys > 1:
            return True
        # n_keys == 1, so first non-key line is 1
        first = self.lines[1].lstrip()
        # multiline block with single key
        # may be property or (deprecated) glyph with plain label
        # we need to check the contents; allows 2-level glyphs only
        return first[:1] in self.inklevels[2] or set(first) == set(self.empty)


##############################################################################
# write file

def globalise_glyph_metrics(glyphs):
    """If all glyph props are equal, take them global."""
    properties = {}
    for key in (
            'shift_up', 'left_bearing', 'right_bearing',
            'shift_left', 'top_bearing', 'bottom_bearing',
            'scalable_width',
        ):
        distinct = set(_g.get_defined(key) for _g in glyphs)
        if len(distinct) == 1:
            value = distinct.pop()
            if value is not None:
                properties[key] = value
    return properties


def transfer_properties(font_props):
    """Transfer properties in order, fix namespaces where needed."""
    props = {}
    for key in FontProperties.__annotations__:
        if key in font_props:
            props[key] = font_props.pop(key)
    # transfer unrecognised properties
    # move non-namespaced unrecognised properties into custom. namespace
    for key, value in font_props.items():
        if '.' in key or key.startswith('_') or key.startswith('-'):
            props[key] = font_props[key]
        else:
            logging.debug(
                "Moving unrecognised property '%s' to '%s' namespace",
                key, CUSTOM_NAMESPACE
            )
            props[f'{CUSTOM_NAMESPACE}.{key}'] = font_props[key]
    return props



def _save_yaff(fonts, outstream):
    """Write fonts to a plaintext stream as yaff."""
    for number, font in enumerate(fonts):
        if len(fonts) > 1:
            outstream.write(BOUNDARY_MARKER + '\n')
        logging.debug('Writing %s to section #%d', font.name, number)
        # write global comment
        if font.get_comment():
            outstream.write(
                format_comment(font.get_comment(), YaffParams.comment)
                + '\n\n'
            )
        # we always output name, font-size and spacing
        # plus anything that is different from the default
        props = {
            'name': font.name,
            'spacing': font.spacing,
        }
        if font.spacing in ('character-cell', 'multi-cell'):
            props['cell_size'] = font.cell_size
        else:
            props['bounding_box'] = font.bounding_box
        if font.levels != 2:
            props['levels'] = font.levels
        # transfer font properties in defined order
        font_props = font.get_properties()
        props |= transfer_properties(font_props)
        global_metrics = globalise_glyph_metrics(font.glyphs)
        # keep only nonzero or non-default globalised properties
        props.update({
            _k: _v for _k, _v in global_metrics.items()
            if _v or _v != Glyph.get_default(_k)
        })
        if props:
            # write recognised yaff properties first, in defined order
            for key, value in props.items():
                if value != font.get_default(key):
                    _write_property(outstream, key, value, font.get_comment(key))
            outstream.write('\n')
        for glyph in font.glyphs:
            _write_glyph(outstream, glyph, global_metrics)

def _write_glyph(outstream, glyph, global_metrics):
    """Write out a single glyph in text format."""
    # glyph comments
    if glyph.comment:
        outstream.write(
            '\n' + format_comment(glyph.comment, YaffParams.comment) + '\n'
        )
    labels = glyph.get_labels() or ['']
    for _label in labels:
        outstream.write(f'{str(_label)}{YaffParams.separator}\n')
    # glyph matrix
    # empty glyphs are stored as 0x0, not 0xm or nx0
    if not glyph.pixels.width or not glyph.pixels.height:
        glyphtxt = f'{YaffParams.tab}{YaffParams.empty}\n'
    else:
        glyphtxt = glyph.pixels.as_text(
            start=YaffParams.tab,
            inklevels=YaffParams.inklevels[glyph.levels],
            end='\n',
        )
    outstream.write(glyphtxt)
    properties = glyph.get_properties()
    for key in global_metrics:
        properties.pop(key, None)
    if properties:
        outstream.write(f'\n')
    for key, value in properties.items():
        if value != glyph.get_default(key):
            _write_property(outstream, key, value, None, indent=YaffParams.tab)
    if properties:
        outstream.write('\n')
    outstream.write('\n')

def _write_property(outstream, key, value, comments, indent=''):
    """Write out a property."""
    if value is None:
        return
    # write property comment
    if comments:
        outstream.write(
            f'\n{indent}{format_comment(comments, YaffParams.comment)}\n'
        )
    key = key.replace('_', '-')
    # write key-value pair
    if isinstance(value, Label) or not isinstance(value, str):
        # do not quote converted non-strings (plus Tag and Char which are str)
        # note that these need special treatment in the reader, or it
        # will strip quotes and misinterpret some labels
        # so non-ascii or single-char tags will be converted to Char not Tag
        quoter = passthrough
    else:
        quoter = _quote_if_needed
    value = str(value)
    if '\n' not in value:
        outstream.write(f'{indent}{key}: {quoter(value)}\n')
    else:
        outstream.write(
            f'{indent}{key}:\n{indent}{YaffParams.tab}' + '{}\n'.format(
                f'\n{indent}{YaffParams.tab}'.join(
                    quoter(_line) for _line in value.splitlines()
                )
            )
        )

def _quote_if_needed(value):
    """See if string value needs double quotes."""
    value = str(value)
    if (
            not value
            or value[0] in YaffParams.quotable
            or value[-1] in YaffParams.quotable
            or value[0] == value[-1] == '"'
            or all(_c in YaffParams.glyphchars for _c in value)
        ):
        return f'"{value}"'
    return value



"""
monobit.storage.fontformats.pilfont - PILfont format

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from pathlib import Path

from monobit.base import safe_import
Image = safe_import('PIL.Image')

from monobit.storage import loaders, savers
from monobit.base import FileFormatError
from monobit.core import Font, Glyph
from monobit.base.struct import big_endian as be
from monobit.render import GlyphMap

from monobit.storage.utils.limitations import ensure_single


_PIL_METRICS = be.Struct(
    dwx='int16',
    dwy='int16',
    dst_x0='int16',
    dst_y0='int16',
    dst_x1='int16',
    dst_y1='int16',
    src_x0='int16',
    src_y0='int16',
    src_x1='int16',
    src_y1='int16',
)


if Image:

    @loaders.register(
        name='pilfont',
        patterns=('*.pil',),
        magic=(b'PILfont\n',),
    )
    def load_pilfont(instream):
        """Load font from PILfont file."""
        signature = instream.readline().strip()
        metadata = instream.readline().strip().split(b';')
        sentinel = instream.readline().strip()
        if signature != b'PILfont':
            raise FileFormatError('PILfont signature not found')
        if sentinel != b'DATA':
            raise FileFormatError('PILfont data sentinel not found')
        try:
            line_height = int(metadata[6])
        except (ValueError, IndexError) as e:
            raise FileFormatError('Malformed PILfont file') from e
        metrics = (
            (_cp, _PIL_METRICS.read_from(instream))
            for _cp in range(256)
        )
        image_name = Path(instream.name).stem + '.pbm'
        with instream.where.open(image_name, 'r') as image_file:
            spritesheet = Image.open(image_file)
            glyphs = []
            for cp, metric in metrics:
                if (
                        not (metric.src_x1 - metric.src_x0)
                        or not (metric.src_y1 - metric.src_y0)
                    ):
                    continue
                crop = spritesheet.crop(
                    (metric.src_x0, metric.src_y0, metric.src_x1, metric.src_y1)
                )
                glyphs.append(
                    Glyph.from_vector(
                        tuple(crop.getdata()),
                        stride=crop.width,
                        inklevels=(0, 255),
                        codepoint=cp,
                        left_bearing=metric.dst_x0,
                        right_bearing=metric.dwx - metric.dst_x1,
                        shift_up=-metric.dst_y1,
                    )
                )
        return Font(glyphs, line_height=line_height)


    @savers.register(linked=load_pilfont)
    def save_pilfont(fonts, outstream, *, max_width:int=800):
        """
        Save font to PILfont file.

        max_width: maximum width of spritesheet
        """
        font = ensure_single(fonts)
        outstream.write(b'PILfont\n')
        # I wonder if the other fields in this header were ever defined
        outstream.write(b';;;;;;%d;\n' % font.line_height)
        outstream.write(b'DATA\n')
        x, y = 0, 0
        glyph_map = GlyphMap()
        for cp in range(256):
            try:
                glyph = font.get_glyph(codepoint=cp, missing='raise')
            except KeyError:
                metrics = _PIL_METRICS()
            else:
                x0, y0 = x, y
                x += glyph.width
                if x > max_width:
                    x, y = 0, y + font.line_height
                    x0, y0 = x, y
                    x = glyph.width
                glyph_map.append_glyph(glyph, x0, y0)
                metrics = _PIL_METRICS(
                    dwx=glyph.advance_width,
                    dwy=0,
                    dst_x0=glyph.left_bearing,
                    dst_y0=-glyph.shift_up - glyph.height,
                    dst_x1=glyph.width + glyph.left_bearing,
                    dst_y1=-glyph.shift_up,
                    src_x0=x0,
                    src_y0=y0,
                    src_x1=x0 + glyph.width,
                    src_y1=y0 + glyph.height,
                )
            # write mapping file
            outstream.write(bytes(metrics))
        # write image
        image_name = Path(outstream.name).stem + '.pbm'
        with outstream.where.open(image_name, 'w') as image_file:
            image = glyph_map.as_image(ink=1, paper=0, border=0, invert_y=True)
            image.save(image_file, format='png')



"""
monobit.storage.containerformats.tarzip - tar and zip containers

(c) 2021--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import io
import logging
import zipfile
import tarfile
import time
from pathlib import Path, PurePosixPath
from itertools import chain

from monobit.base import FileFormatError
from ..containers import Archive
from ..base import containers
from ..streams import KeepOpen, Stream
from ..magic import Magic


class ZipTarBase(Archive):

    def __init__(self, file, mode='r'):
        """Create wrapper."""
        # reading zipfile needs a seekable stream, drain to buffer if needed
        self._stream = Stream(file, mode)
        # create the zipfile
        self._archive = self._create_archive(self._stream, mode)
        # output files, to be written on close
        self._files = []
        super().__init__(file, mode)

    def close(self):
        """Close the archive, ignoring errors."""
        if self.mode == 'w' and not self.closed:
            for file in self._files:
                logging.debug("Writing out '%s' to archive %s.", file.name, self)
                self._write_out(file)
                file.close()
        try:
            self._archive.close()
        except EnvironmentError as e:
            # e.g. BrokenPipeError
            logging.debug(e)
        self._stream.close()
        super().close()

    def decode(self, name, **kwargs):
        """Open a stream in the container."""
        filename = str(PurePosixPath(self.root) / name)
        # always open as binary
        logging.debug(
            "Opening readable stream '%s' on container %s.",
            name, self
        )
        try:
            file = self._open_read(filename, **kwargs)
        except KeyError:
            raise FileNotFoundError(f"'{filename}' not found on {self}")
        if file is None:
            raise IsADirectoryError(f"'{filename}' is a directory")
        # .name is not writeable, so we need to wrap
        return Stream(file, mode='r', name=name)

    def encode(self, name, **kwargs):
        """Open a stream for writing to the container."""
        filename = str(PurePosixPath(self.root) / name)
        # always open as binary
        logging.debug(
            "Opening writeable stream '%s' on container %s.",
            name, self
        )
        # stop BytesIO from being closed until we want it to be
        newfile = Stream(KeepOpen(io.BytesIO()), mode='w', name=name)
        if any(name == _file.name for _file in self._files):
            logging.warning("Creating multiple files of the same name '%s'.", name)
        self._files.append(newfile)
        return newfile


###############################################################################

@containers.register(
    name='zip',
    magic=(b'PK\x03\x04',),
    patterns=('*.zip',),
)
class ZipContainer(ZipTarBase):
    """Zip-file wrapper."""

    def decode(self, name, *, password:bytes=None):
        """
        Extract file from zip archive.

        password: password for encrypted archive entry.
        """
        return super().decode(name, password=password)

    def encode(self, name):
        """Store file in zip archive."""
        return super().encode(name)

    def list(self):
        """List full contents of archive."""
        # construct directory entries even if they are missing from the zip
        ziplist = tuple(set(chain(*(
            # the list() is only needed for python 3.9
            (_name, *(f'{_path}/' for _path in list(Path(_name).parents)[:-1]))
            for _name in self._archive.namelist()
        ))))
        ziplist += tuple(str(PurePosixPath(self.root) / _file.name) for _file in self._files)
        return ziplist

    @staticmethod
    def _create_archive(stream, mode):
        try:
            return zipfile.ZipFile(
                stream, mode,
                compression=zipfile.ZIP_DEFLATED
            )
        except zipfile.BadZipFile as exc:
            raise FileFormatError(exc) from exc

    def _open_read(self, filename, *, password):
        filename = filename.removesuffix('/')
        try:
            return self._archive.open(filename, 'r', pwd=password)
        except KeyError:
            # return None for open() on directories (like tarfile does)
            if filename + '/' in self.list():
                return None
            raise

    def _write_out(self, file):
        bytearray = file.getvalue()
        self._archive.writestr(
            str(PurePosixPath(self.root) / file.name), bytearray
        )



###############################################################################

@containers.register(
    name='tar',
    # maybe
    magic=(
        Magic.offset(257) + b'ustar',
    ),
    patterns=('*.tar',),
)
class TarContainer(ZipTarBase):
    """Tar-file wrapper."""

    def decode(self, name):
        """
        Extract file from tar archive.

        password: password for encrypted archive entry.
        """
        return super().decode(name)

    def encode(self, name):
        """Store file in tar archive."""
        return super().encode(name)

    def list(self):
        """List full contents of archive."""
        tarlist = tuple(
            f'{_name}/'
            if self._archive.getmember(_name).isdir()
            else _name
            for _name in self._archive.getnames()
        )
        tarlist += tuple(str(PurePosixPath(self.root) / _file.name) for _file in self._files)
        return tarlist

    @staticmethod
    def _create_archive(stream, mode):
        try:
            return tarfile.open(fileobj=stream, mode=mode)
        except tarfile.ReadError as exc:
            raise FileFormatError(exc) from exc

    def _open_read(self, filename):
        return self._archive.extractfile(filename)

    def _write_out(self, file):
        name = file.name
        tinfo = tarfile.TarInfo(str(PurePosixPath(self.root) / name))
        tinfo.mtime = time.time()
        tinfo.size = len(file.getvalue())
        file.seek(0)
        self._archive.addfile(tinfo, file)



"""
monobit.storage.fontformats.fon.windows - Windows FON and FNT files

`monobit.storage.fontformats.fon.windows` is copyright 2019--2024 Rob Hagemans
`mkwinfont` is copyright 2001 Simon Tatham. All rights reserved.
`dewinfont` is copyright 2001,2017 Simon Tatham. All rights reserved.

See `LICENSE.md` in this package's directory.
"""

from . import fnt, ne, pe



"""
monobit.storage.containerformats.macbinary - MacBinary containers

(c) 2021--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.base.struct import big_endian as be
from monobit.base.binary import align

from monobit.base import FileFormatError
from ..streams import Stream
from ..magic import Magic
from ..base import containers
from ..containers import FlatFilterContainer


##############################################################################
# MacBinary container
# v1: https://www.cryer.co.uk/file-types/b/bin_/original_mac_binary_format_proposal.htm
# v2: https://files.stairways.com/other/macbinaryii-standard-info.txt
# v2 defines additional fields inside an area zeroed in v1. we can ignore them.

_MACBINARY_HEADER = be.Struct(
    # Offset 000-Byte, old version number, must be kept at zero for compatibility
    old_version='byte',
    # Offset 001-Byte, Length of filename (must be in the range 1-63)
    filename_length='byte',
    # Offset 002-1 to 63 chars, filename (only "length" bytes are significant).
    filename='63s',
    # Offset 065-Long Word, file type (normally expressed as four characters)
    file_type='4s',
    # Offset 069-Long Word, file creator (normally expressed as four characters)
    file_creator='4s',
    # Offset 073-Byte, original Finder flags
    original_finder_flags='byte',
    # Offset 074-Byte, zero fill, must be zero for compatibility
    zero_0='byte',
    # Offset 075-Word, file's vertical position within its window.
    window_vert='word',
    # Offset 077-Word, file's horizontal position within its window.
    window_horiz='word',
    # Offset 079-Word, file's window or folder ID.
    window_id='word',
    # Offset 081-Byte, "Protected" flag (in low order bit).
    protected='byte',
    # Offset 082-Byte, zero fill, must be zero for compatibility
    zero_1='byte',
    # Offset 083-Long Word, Data Fork length (bytes, zero if no Data Fork).
    data_length='dword',
    # Offset 087-Long Word, Resource Fork length (bytes, zero if no R.F.).
    rsrc_length='dword',
    # Offset 091-Long Word, File's creation date
    creation_date='dword',
    # Offset 095-Long Word, File's "last modified" date.
    last_modified_date='dword',
    # Offset 099-Word, length of Get Info comment to be sent after the resource
    # fork (if implemented, see below).
    get_info_length='word',
    # *Offset 101-Byte, Finder Flags, bits 0-7. (Bits 8-15 are already in byte 73)
    finder_flags='byte',
    # *Offset 116-Long Word, Length of total files when packed files are unpacked.
    packed_length='dword',
    # *Offset 120-Word, Length of a secondary header.  If this is non-zero,
    #              Skip this many bytes (rounded up to the next multiple of 128)
    #              This is for future expansion only, when sending files with
    #              MacBinary, this word should be zero.
    second_header_length='dword',
    # *Offset 122-Byte, Version number of Macbinary II that the uploading program
    # is written for (the version begins at 129)
    writer_version='byte',
    # *Offset 123-Byte, Minimum MacBinary II version needed to read this file
    # (start this value at 129 129)
    reader_version='byte',
    # *Offset 124-Word, CRC of previous 124 bytes
    crc='word',
    # from v1 desc:
    # > 126 2 Reserved for computer type and OS ID
    # > (this field will be zero for the current Macintosh).
    reserved='word',
    # *This is newly defined for MacBinary II.
)


@containers.register(
    name='macbin',
    magic=(
        # FFILDMOV is a maybe
        Magic.offset(65) + b'FFILDMOV',
    ),
)
class MacBinary(FlatFilterContainer):
    """MacBinary container."""

    def decode(self, name):
        """
        Decode data and resource fork from MacBinary container.
        """
        return super().decode(name)

    def encode(self, name):
        """
        Writing to MacBinary is not supported.
        """
        raise ValueError(
            'Writing to MacBinary is not supported.'
        )

    def decode_all(self, stream):
        """Parse a MacBinary file."""
        data = stream.read()
        header = _MACBINARY_HEADER.from_bytes(data)
        ofs = 128
        if header.old_version != 0:
            raise FileFormatError(
                'Not a MacBinary file: incorrect version field'
                f' ({header.old_version}).'
            )
        if header.writer_version > 128:
            ofs += align(header.second_header_length, 7)
        data_fork = data[ofs:ofs+header.data_length]
        ofs += align(header.data_length, 7)
        rsrc_fork = data[ofs:ofs+header.rsrc_length]
        name = header.filename.decode('mac-roman').strip()
        if name:
            return {
                f'data/{name}': data_fork,
                f'rsrc/{name}': rsrc_fork,
            }
        else:
            return {
                f'data': data_fork,
                f'rsrc': rsrc_fork,
            }



"""
monobit.storage.fontformats.dashen - Dashen font

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging

from monobit.base.binary import ceildiv
from monobit.base.struct import little_endian as le
from monobit.base import Props
from monobit.storage import loaders, savers, Magic
from monobit.core import Font, Glyph, Raster

from .softfont.pcl import load_hppcl


@loaders.register(
    name='dashen',
    patterns=('*.pft',),
    # maybe-magic. first entry in offsets table, for NUL
    magic=(Magic.offset(14) + b'\xff\xff\xff\xff',),
)
def load_dashen(instream):
    """Load font from Dashen .pft file."""
    dashen_data = _read_dashen(instream)
    # screen font ?
    screen = _convert_dashen(*dashen_data)
    # PCL printer font
    if instream.peek(1):
        printer = load_hppcl(instream)
        printer = printer.modify(encoding='dashen').label()
        return screen, printer
    else:
        return screen


# 14 bytes. all guesses
_DASHEN_HEADER = le.Struct(
    # uint32 offset or size?
    unknown0=le.uint8*4,
    line_height='uint16',
    maybe_cell_width='uint16',
    # 4 (includes pcl font) 3 (no pcl) 1 (1-byte glyph header
    maybe_format_version='uint16',
    maybe_left='uint8',
    maybe_right='uint8',
    top='uint8',
    # descent plus leading?
    bottom='uint8',
)


_DASHEN_GLYPH_HEADER = le.Struct(
    width='uint8',
    shift='uint8',
    height='uint8',
)

def _read_dashen(instream):
    """Read Dashen screen font."""
    header = _DASHEN_HEADER.read_from(instream)
    logging.debug(header)
    # offsets from start of glyph table (i.e. from 1070)
    # 255 le.int32 values, -1 for unused
    offsets = (le.int32 * 256).read_from(instream)
    # perhaps a name string? would be in encoding
    unknown2 = instream.read(32)
    logging.debug(unknown2)
    glyphdata = []
    headers = []
    logging.debug(offsets)
    for offset in offsets:
        if offset == -1:
            glyphdata.append(None)
            headers.append(None)
            continue
        ghdr = _DASHEN_GLYPH_HEADER.read_from(instream)
        headers.append(ghdr)
        glyphdata.append(tuple(
            instream.read(ghdr.width)
            for _ in range(ceildiv(ghdr.height, 8))
        ))
    if instream.peek(1):
        codepoints = (le.uint8 * 256).read_from(instream)
    else:
        codepoints = range(256)
    props = Props(**vars(header), unknown2=unknown2)
    return props, glyphdata, headers, codepoints


def _convert_dashen(props, glyphdata, headers, codepoints):
    """Convert from Dashen font data to monobit."""
    glyphs = []
    for gdata, ghdr, codepoint in zip(glyphdata, headers, codepoints):
        if gdata is None:
            continue
        raster = Raster.concatenate(*(
            Raster.from_bytes(_bytes, width=8)
            for _bytes in gdata
        ))
        raster = raster.transpose()
        # clip off the alignment bits
        raster = raster.crop(bottom=raster.height-ghdr.height)
        glyphs.append(Glyph(
            raster, codepoint=codepoint,
            shift_up=props.top-raster.height-ghdr.shift+1
        ))
    font = Font(
        glyphs,
        line_height=props.line_height,
        encoding='dashen',
    )
    return font.label()



"""
monobit.storage.fontformats.softfont.wyse - Wyse-60 soft font

(c) 2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

# http://bitsavers.org/pdf/wyse/WY-60/880261-01A_WY-60_Programmers_Guide_Jan87.pdf
# pp 48--52

import logging

from monobit.storage import loaders, savers
from monobit.core import Font, Raster, Glyph
from monobit.base import Coord, FileFormatError

from monobit.storage.utils.limitations import ensure_single, ensure_charcell


_WYSE_ESC_SEQ = b'\x1bcA'


@loaders.register(
    name='wyse',
    # maybe
    magic=(_WYSE_ESC_SEQ,)
)
def load_wyse(instream):
    """Load character-cell fonts from Wyse-60 soft font file."""
    glyphs = []
    while True:
        tri = instream.peek(3)
        if len(tri) < 3:
            break
        if tri[:3] == _WYSE_ESC_SEQ:
            glyphs.append(_read_wyse_glyph(instream))
        else:
            instream.read(1)
    if not glyphs:
        raise FileFormatError(
            'Not a Wyse-60 soft font: ESC c A sequence not found.'
        )
    return Font(glyphs)


def _read_wyse_glyph(instream):
    """Read a single-glyph escape sequence."""
    esc_seq = instream.read(3)
    try:
        bank = int(instream.read(1))
        codepoint, = bytes.fromhex(instream.read(2).decode('ascii'))
        glyph = Glyph.from_hex(
            instream.read(32).decode('ascii'), width=8,
            # interpret bank value (first char) as part of codepoint
            # alternatively, we could define 3 fonts, one per bank
            codepoint=bank * 0x100 + codepoint,
        )
        # rightmost column should be repeated
        glyph = glyph.expand(right=2)
    except (ValueError, UnicodeError) as e:
        raise FileFormatError(
            'Not a Wyse-60 soft font: malformed character definition.'
        ) from e
    ctrl_y = instream.read(1)
    return glyph


@savers.register(linked=load_wyse)
def save_wyse(fonts, outstream):
    """Save character-cell fonts to Wyse-60 soft font file."""
    font = ensure_single(fonts)
    font = ensure_charcell(font)
    if (font.cell_size.x > 8 or font.cell_size.y > 16):
        raise ValueError(
            'This format can only store character-cell fonts '
            'of cell-size 8x16 or smaller.'
        )
    font = font.label(codepoint_from=font.encoding)
    for glyph in font.glyphs:
        _write_wyse_glyph(glyph, outstream)


def _write_wyse_glyph(glyph, outstream):
    """Write a single-glyph escape sequence."""
    glyph = glyph.expand(right=8-glyph.width, bottom=16-glyph.height)
    if glyph.codepoint and int(glyph.codepoint) <= 0x3ff:
        outstream.write(_WYSE_ESC_SEQ)
        outstream.write(b'%03X' % int(glyph.codepoint))
        outstream.write(glyph.as_hex().upper().encode('ascii'))
        outstream.write(b'\x19')
    else:
        logging.warning(
            'Cannot encode codepoints > 0x3ff. '
            f'Skipping {glyph.codepoint}.'
        )



"""
monobit.plumbing.scripting - scripting utilities

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

import logging
from functools import wraps, partial

from ..base import Any, passthrough, to_int, CONVERTERS, NOT_SET


class ArgumentError(TypeError):
    """Invalid keyword argument."""

    def __init__(self, func, arg):
        super().__init__(f'{arg} is an invalid keyword for {func}')


###############################################################################
# mark functions for scripting
# annotations give converters from string to desired type
# docstings provide help text

scriptables = {}


def scriptable(
        *args, script_args=None,
        record=True, pack_operation=False, wrapper=False,
    ):
    """
    Decorator to register operation for scripting.

    Decorated functions get
    - script arguments in annotations
    - automatic type conversion

    script_args: additional arguments not given in annotations
    pack_operation: function works on sequence of fonts
    wrapper: enable keyword argument passthrough
    """
    if not args:
        # called as @scriptable(script_args=...)
        # return decorator with these arguments set as extra args
        return partial(
            scriptable, script_args=script_args,
            record=record, pack_operation=pack_operation,
            wrapper=wrapper,
        )
    # called as @scriptable
    func, = args

    @wraps(func)
    def _scriptable_func(*args, **kwargs):
        return func(*args, **kwargs)

    _scriptable_func.__annotations__.update(script_args or {})
    _scriptable_func = convert_arguments(_scriptable_func)
    if not wrapper:
        _scriptable_func = check_arguments(_scriptable_func)
    _scriptable_func.pack_operation = pack_operation
    # register the scriptable function
    scriptables[_scriptable_func.__name__] = _scriptable_func
    return _scriptable_func


def check_arguments(func):
    """Check if arguments to function are in the registered script args."""

    @wraps(func)
    def _checked_func(*args, **kwargs):
        for kwarg in kwargs:
            if kwarg not in func.__annotations__:
                raise ArgumentError(func.__name__, kwarg) from None
        # call wrapped function
        return func(*args, **kwargs)

    return _checked_func


def convert_arguments(func):
    """Convert given arguments to the type in registered script args."""

    @wraps(func)
    def _converted_func(*args, **kwargs):
        # apply converters to argument
        conv_kwargs = {}
        for kwarg, value in kwargs.items():
            # skip not-specified arguments
            if value is NOT_SET:
                continue
            _type = func.__annotations__.get(kwarg, Any)
            converter = CONVERTERS.get(_type, _type)
            conv_kwargs[kwarg] = converter(value)
        # call wrapped function
        result = func(*args, **conv_kwargs)
        return result

    return _converted_func



def manage_arguments(loader):
    loader = convert_arguments(loader)
    loader = check_arguments(loader)
    return loader


def take_arguments(func, argdict):
    """Subset the argument dictionary with args that occur in registered script args."""
    if not argdict:
        return {}
    _types = {
        _key: func.__annotations__.get(_key, Any)
        for _key in func.__annotations__
    }
    converters = {
        _key: CONVERTERS.get(_type, _type)
        for _key, _type in _types.items()
    }
    kwargs = {
        _key: converters[_key](_value) for _key, _value in argdict.items()
        if _key in converters
    }
    return kwargs



"""
monobit.encoding.base - base classes and functions for encoding

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""



class NotFoundError(KeyError):
    """Encoding not found."""


class EncodingName(str):

    # replacement patterns for normalisation
    # longest first to avoid partial match
    _patterns = {
        'microsoftcp': 'windows',
        'microsoft': 'windows',
        'msdoscp': 'oem',
        'oemcp': 'oem',
        'msdos': 'oem',
        'ibmcp': 'ibm',
        'apple': 'mac',
        'macos': 'mac',
        'doscp': 'oem',
        'mscp': 'windows',
        'dos': 'oem',
        'pc': 'oem',
        'ms': 'windows',
        # mac-roman also known as x-mac-roman etc.
        'x': '',
    }

    def __new__(cls, value=''):
        """Convert value to encoding name."""
        value = cls._normalise_name(str(value))
        return super().__new__(cls, value)

    def __eq__(self, other):
        """Check if two names match."""
        return self._normalise_for_match() == EncodingName(other)._normalise_for_match()

    def __hash__(self):
        return str.__hash__(self._normalise_for_match())

    @staticmethod
    def _normalise_name(name=''):
        """Replace encoding name with normalised variant for display."""
        return name.lower().replace('_', '-').replace(' ', '-')

    def _normalise_for_match(self):
        """Further normalise names to base form."""
        name = str(self)
        # remove underscores, spaces, dashes and dots
        for char in '-.':
            name = name.replace(char, '')
        # try replacements
        for start, replacement in self._patterns.items():
            if name.startswith(start):
                name = replacement + name[len(start):]
                break
        return name



class Encoder:
    """
    Convert between unicode, ordinals and tags.
    Encoder objects act on single-glyph codes only, which may be single- or multi-codepoint.
    They need not encode/decode between full strings and bytes.
    """

    def __init__(self, name):
        """Set encoder name."""
        self.name = EncodingName(name)

    def char(self, *labels):
        """Convert labels to character, return empty string if missing."""
        raise NotImplementedError

    def codepoint(self, *labels):
        """Convert labels to codepoint, return None if missing."""
        raise NotImplementedError

    def tag(self, *labels):
        """Convert labels to tag, return None if missing."""
        raise NotImplementedError

    def __repr__(self):
        """Representation."""
        return f"{type(self).__name__}(name='{self.name}')"


class EncoderBuilder:
    """Callable returning an Encoder and allowing for lazy composition."""

    def __init__(self, callable):
        self._callable = callable

    def __call__(self):
        return self._callable()

    # delayed operations

    def __or__(self, other):
        """Return encoding overlaid with all characters defined in right-hand side."""
        if isinstance(other, Encoder):
            def delayed_or():
                return self() | other
        else:
            def delayed_or():
                return self() | other()
        return EncoderBuilder(delayed_or)

    def subset(self, codepoint_range):
        """Return encoding only for given range of codepoints."""
        def delayed_subset():
            return self().subset(codepoint_range)
        return EncoderBuilder(delayed_subset)


# registry of charmap/tagmap file format readers
encoding_readers = {}

def register_reader(format, **default_kwargs):
    """Decorator to register charmap reader."""
    def decorator(reader):
        encoding_readers[format] = (reader, default_kwargs)
        return reader
    return decorator
